from enum import Enum

from ..qtcompat.QtCore import QSize, Qt, Signal
from ..qtcompat.QtGui import QFontMetrics, QValidator
from ..qtcompat.QtWidgets import QAbstractSpinBox, QStyle, QStyleOptionSpinBox


class _EmitPolicy(Enum):
    EmitIfChanged = 0
    AlwaysEmit = 1
    NeverEmit = 2


class _AnyIntValidator(QValidator):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)

    def validate(self, input: str, pos: int):
        if not input.lstrip("-"):
            return QValidator.State.Intermediate, input, len(input)
        if input.lstrip("-").isnumeric():
            return QValidator.State.Acceptable, input, len(input)
        return QValidator.State.Invalid, input, len(input)


class QLargeIntSpinBox(QAbstractSpinBox):
    """An integer spinboxes backed by unbound python integer

    Qt's built-in ``QSpinBox`` is backed by a signed 32-bit integer.
    This could become limiting, particularly in large dense segmentations.
    This class behaves like a ``QSpinBox`` backed by an unbound python int.

    Does not yet support "prefix", "suffix" or "specialValue" like QSpinBox.
    """

    textChanged = Signal(str)
    valueChanged = Signal(object)  # object instead of int for large ints

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._value: int = 0
        self._minimum: int = 0
        self._maximum: int = 2 ** 64 - 1
        self._single_step: int = 1
        self._pending_emit = False
        validator = _AnyIntValidator(self)
        self.lineEdit().setValidator(validator)
        self.lineEdit().textChanged.connect(self._editor_text_changed)
        self.setValue(0)

    # ###############  Public Functions  #######################

    def value(self):
        return self._value

    def setValue(self, value):
        self._setValue(value, _EmitPolicy.EmitIfChanged)

    def minimum(self):
        return self._minimum

    def setMinimum(self, min):
        self._minimum = int(min)

    def maximum(self):
        return self._maximum

    def setMaximum(self, max):
        self._maximum = int(max)

    def setRange(self, minimum, maximum):
        self.setMinimum(minimum)
        self.setMaximum(maximum)

    def singleStep(self):
        return self._single_step

    def setSingleStep(self, step):
        self._single_step = int(step)

    # TODO: add prefix/suffix/stepType

    # ###############  QtOverrides  #######################

    def focusOutEvent(self, e) -> None:
        if self._pending_emit:
            self._interpret(_EmitPolicy.EmitIfChanged)
        return super().focusOutEvent(e)

    def closeEvent(self, e) -> None:
        if self._pending_emit:
            self._interpret(_EmitPolicy.EmitIfChanged)
        return super().closeEvent(e)

    def keyPressEvent(self, e) -> None:
        if e.key() in (Qt.Key.Key_Enter, Qt.Key.Key_Return):
            self._interpret(
                _EmitPolicy.AlwaysEmit
                if self.keyboardTracking()
                else _EmitPolicy.EmitIfChanged
            )
        return super().keyPressEvent(e)

    def stepBy(self, steps: int) -> None:
        step = self._single_step
        old = self._value
        e = _EmitPolicy.EmitIfChanged
        if self._pending_emit:
            self._interpret(_EmitPolicy.NeverEmit)
            if self._value != old:
                e = _EmitPolicy.AlwaysEmit
        self._setValue(self._bound(self._value + (step * steps)), e)

    def stepEnabled(self):
        flags = QAbstractSpinBox.StepEnabledFlag.StepNone
        if self.isReadOnly():
            return flags
        if self._value < self._maximum:
            flags |= QAbstractSpinBox.StepEnabledFlag.StepUpEnabled
        if self._value > self._minimum:
            flags |= QAbstractSpinBox.StepEnabledFlag.StepDownEnabled
        return flags

    def sizeHint(self):
        self.ensurePolished()
        fm = QFontMetrics(self.font())
        h = self.lineEdit().sizeHint().height()
        if hasattr(fm, "horizontalAdvance"):
            # Qt >= 5.11
            w = fm.horizontalAdvance(str(self._value)) + 3
        else:
            w = fm.width(str(self._value)) + 3
        w = max(36, w)
        opt = QStyleOptionSpinBox()
        self.initStyleOption(opt)
        hint = QSize(w, h)
        return self.style().sizeFromContents(
            QStyle.ContentsType.CT_SpinBox, opt, hint, self
        )

    # ###############  Implementation Details  #######################

    def _setValue(self, value, policy):
        self._value, old = self._bound(int(value)), self._value
        self._pending_emit = False
        self._updateEdit()
        self.update()

        if policy is _EmitPolicy.AlwaysEmit or (
            policy is _EmitPolicy.EmitIfChanged and self._value != old
        ):
            self._pending_emit = False
            self.textChanged.emit(self.lineEdit().displayText())
            self.valueChanged.emit(self._value)

    def _updateEdit(self):
        new_text = str(self._value)
        if self.lineEdit().text() == new_text:
            return
        self.lineEdit().setText(new_text)

    def _interpret(self, policy):
        text = self.lineEdit().displayText() or str(self._value)
        v = int(text)
        self._setValue(v, policy)

    def _editor_text_changed(self, t):
        if self.keyboardTracking():
            self._setValue(int(t), _EmitPolicy.EmitIfChanged)
            self.lineEdit().setFocus()
            self._pending_emit = False
        else:
            self._pending_emit = True

    def _bound(self, value):
        return max(self._minimum, min(self._maximum, value))



"""superqt is a collection of QtWidgets for python."""
try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"


from ._eliding_label import QElidingLabel
from .collapsible import QCollapsible
from .combobox import QEnumComboBox
from .sliders import (
    QDoubleRangeSlider,
    QDoubleSlider,
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
    QLabeledRangeSlider,
    QLabeledSlider,
    QRangeSlider,
)
from .spinbox import QLargeIntSpinBox
from .utils import QMessageHandler, ensure_main_thread, ensure_object_thread

__all__ = [
    "ensure_main_thread",
    "ensure_object_thread",
    "QDoubleRangeSlider",
    "QDoubleSlider",
    "QElidingLabel",
    "QLabeledDoubleRangeSlider",
    "QLabeledDoubleSlider",
    "QLabeledRangeSlider",
    "QLabeledSlider",
    "QLargeIntSpinBox",
    "QMessageHandler",
    "QRangeSlider",
    "QEnumComboBox",
    "QCollapsible",
]



import logging
from contextlib import suppress
from typing import List, NamedTuple, Optional

from superqt.qtcompat.QtCore import (
    QMessageLogContext,
    QtMsgType,
    qInstallMessageHandler,
)


class Record(NamedTuple):
    level: int
    message: str
    ctx: dict


class QMessageHandler:
    """A context manager to intercept messages from Qt.

    Parameters
    ----------
    logger : logging.Logger, optional
        If provided, intercepted messages will be logged with `logger` at the
        corresponding python log level, by default None

    Attributes
    ----------
    records: list of tuple
        Captured messages. This is a 3-tuple of:
        `(log_level: int, message: str, context: dict)`

    Examples
    --------

    >>> handler = QMessageHandler()
    >>> handler.install()  # now all Qt output will be available at mh.records

    >>> with QMessageHandler() as handler:  # temporarily install
    ...     ...

    >>> logger = logging.getLogger(__name__)
    >>> with QMessageHandler(logger):  # re-reoute Qt messages to a python logger.
    ...    ...
    """

    _qt2loggertype = {
        QtMsgType.QtDebugMsg: logging.DEBUG,
        QtMsgType.QtInfoMsg: logging.INFO,
        QtMsgType.QtWarningMsg: logging.WARNING,
        QtMsgType.QtCriticalMsg: logging.ERROR,  # note
        QtMsgType.QtFatalMsg: logging.CRITICAL,  # note
        QtMsgType.QtSystemMsg: logging.CRITICAL,
    }

    def __init__(self, logger: Optional[logging.Logger] = None):
        self.records: List[Record] = []
        self._logger = logger
        self._previous_handler: Optional[object] = "__uninstalled__"

    def install(self):
        """Install this handler (override the current QtMessageHandler)."""
        self._previous_handler = qInstallMessageHandler(self)

    def uninstall(self):
        """Uninstall this handler, restoring the previous handler."""
        if self._previous_handler != "__uninstalled__":
            qInstallMessageHandler(self._previous_handler)

    def __repr__(self):
        n = type(self).__name__
        return f"<{n} object at {hex(id(self))} with {len(self.records)} records>"

    def __enter__(self):
        """Enter a context with this handler installed"""
        self.install()
        return self

    def __exit__(self, *args):
        self.uninstall()

    def __call__(self, msgtype: QtMsgType, context: QMessageLogContext, message: str):
        level = self._qt2loggertype[msgtype]

        # PyQt seems to throw an error if these are simply empty
        ctx = dict.fromkeys(["category", "file", "function", "line"])
        with suppress(UnicodeDecodeError):
            ctx["category"] = context.category
        with suppress(UnicodeDecodeError):
            ctx["file"] = context.file
        with suppress(UnicodeDecodeError):
            ctx["function"] = context.function
        with suppress(UnicodeDecodeError):
            ctx["line"] = context.line

        self.records.append(Record(level, message, ctx))
        if self._logger is not None:
            self._logger.log(level, message, extra=ctx)



from __future__ import annotations

import os
import sys
import warnings
from importlib import abc, import_module, util
from typing import TYPE_CHECKING, Optional, Sequence, Union

if TYPE_CHECKING:
    from importlib.machinery import ModuleSpec
    from types import ModuleType


class QtMissingError(ImportError):
    """Error raise if no bindings could be selected."""


VALID_APIS = {
    "pyqt5": "PyQt5",
    "pyqt6": "PyQt6",
    "pyside2": "PySide2",
    "pyside6": "PySide6",
}

# Detecting if a binding was specified by the user
_requested_api = os.getenv("QT_API", "").lower()
_forced_api = os.getenv("FORCE_QT_API")

# warn if an invalid API has been requested
if _requested_api and _requested_api not in VALID_APIS:
    warnings.warn(
        f"invalid QT_API specified: {_requested_api}. "
        f"Valid values include {set(VALID_APIS)}"
    )
    _forced_api = None
    _requested_api = ""

# TODO: FORCE_QT_API requires also using QT_API ... does that make sense?

# now we'll try to import QtCore
_QtCore: Optional[ModuleType] = None

# If `FORCE_QT_API` is not set, we first look for previously imported bindings
if not _forced_api:
    for api_name, module_name in VALID_APIS.items():
        if module_name in sys.modules:
            _QtCore = import_module(f"{module_name}.QtCore")
            break

if _QtCore is None:
    # try the requested API first, and if _forced_api is True,
    # raise an ImportError if it doesn't work.
    # Otherwise go through the list of Valid APIs until something imports
    requested = VALID_APIS.get(_requested_api)
    for module_name in sorted(VALID_APIS.values(), key=lambda x: x != requested):
        try:
            _QtCore = import_module(f"{module_name}.QtCore")
            break
        except ImportError:
            if _forced_api:
                ImportError(
                    "FORCE_QT_API set and unable to import requested QT_API: {e}"
                )

# didn't find one...  not going to work
if _QtCore is None:
    raise QtMissingError(f"No QtCore could be found. Tried: {VALID_APIS.values()}")

# load variables based on what we found.
if not _QtCore.__package__:
    raise RuntimeError("QtCore does not declare __package__?")

API_NAME = _QtCore.__package__
PYSIDE2 = API_NAME == "PySide2"
PYSIDE6 = API_NAME == "PySide6"
PYQT5 = API_NAME == "PyQt5"
PYQT6 = API_NAME == "PyQt6"
QT_VERSION = getattr(_QtCore, "QT_VERSION_STR", "") or getattr(_QtCore, "__version__")

# lastly, emit a warning if we ended up with an API other than the one requested
if _requested_api and API_NAME != VALID_APIS[_requested_api]:
    warnings.warn(
        f"Selected binding {_requested_api!r} could not be found, using {API_NAME!r}"
    )


# Setup the meta path finder that lets us import anything using `superqt.qtcompat.Mod`
class SuperQtImporter(abc.MetaPathFinder):
    def find_spec(
        self,
        fullname: str,
        path: Optional[Sequence[Union[bytes, str]]],
        target: Optional[ModuleType] = None,
    ) -> Optional[ModuleSpec]:
        """Find a spec for the specified module.

        If fullname is superqt.X or superqt.qtcompat.Xx ...
        it will look for API_NAME.X instead...

        See https://docs.python.org/3/reference/import.html#the-meta-path
        """
        if fullname.startswith(__name__):
            spec = fullname.replace(__name__, API_NAME)
            return util.find_spec(spec)
        return None


def _get_qtmodule(mod_name: str) -> ModuleType:
    """Convenience to get a submodule from the current QT_API"""
    _mod_name = mod_name.rsplit(".", maxsplit=1)[-1]
    return import_module(f"{API_NAME}.{_mod_name}")


sys.meta_path.append(SuperQtImporter())



__all__ = (
    "create_worker",
    "ensure_main_thread",
    "ensure_object_thread",
    "FunctionWorker",
    "GeneratorWorker",
    "new_worker_qthread",
    "QMessageHandler",
    "thread_worker",
    "WorkerBase",
)

from ._ensure_thread import ensure_main_thread, ensure_object_thread
from ._message_handler import QMessageHandler
from ._qthreading import (
    FunctionWorker,
    GeneratorWorker,
    WorkerBase,
    create_worker,
    new_worker_qthread,
    thread_worker,
)
